# -*- coding: utf-8 -*-
#
# MONK automated test framework
#
# Copyright (C) 2013 DResearch Fahrzeugelektronik GmbH
# Written and maintained by MONK Developers <project-monk@dresearch-fe.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version
# 3 of the License, or (at your option) any later version.
#

"""
Instead of creating :py:class:`~monk_tf.dev.Device` and
:py:class:`~monk_tf.conn.AConnection` objects by yourself, you can also choose
to put corresponding data in a separate file and let this layer handle the
object concstruction and destruction for you. Doing this will probably make
your test code look more clean, keep the number of places where you need to
change something as small as possible, and lets you reuse data that you already
have described.

A hello world test with it looks like this::

    import nose
    from monk_tf import fixture

    def test_hello():
        ''' say hello
        '''
        # set up
        h = fixture.Fixture('target_device.cfg')
        expected_out = "hello"
        # execute
        out = h.devs[0].cmd('echo "hello"')
        # assert
        nose.tools.eq_(expected_out, out)
        # tear down
        h.tear_down()

When using this layer setting up a device only takes one line of code. The rest
of the information is in the ``target_device.cfg`` file. :term:`MONK` currently 
comes with one text format parser predefined, which is the
:py:class:`~monk_tf.fixture.XiniParser`. ``Xini`` is short for
:term:`extended INI`. You may, however, use any data format you want, if you
extend the :py:class:`~monk_tf.fixture.AParser` class accordingly.

An example ``Xini`` data file might look like this::

    [device1]
        type=Device
        [[serial1]]
            type=SerialConnection
            port=/dev/ttyUSB1
            user=example
            password=secret

As you can see it looks like an :term:`INI` file. There are sections,
consisting of a title enclosed in squared brackets (``[]``) and lists of
properties, consisting of key-value pairs separated by equality signs (``=``).
The unusual part is that the section *serial1* is surrounded by two pairs of
squared brackets (``[]``). This is the specialty of this format indicating that
*serial1* is a subsection of *device1* and therefore is a nested section. This
nesting can be done unlimited, by surrounding a section with more and more
pairs of squared brackets (``[]``) according to the level of nesting intended.
In this example *serial1* belongs to *device1* and the types indicate the
corresponding :term:`MONK` object to be created.

Classes
-------
"""

import os
import os.path as op
import sys
import logging
import collections

import configobj as config

import conn
import dev

logger = logging.getLogger(__name__)

############
#
# Exceptions
#
############

class AFixtureException(Exception):
    """ Base class for exceptions of the fixture layer.

    If you want to make sure that you catch all exceptions that are related
    to this layer, you should catch *AFixtureExceptions*. This also means
    that if you extend this list of exceptions you should inherit from this
    exception and not from :py:exc:`~exceptions.Exception`.
    """
    pass

class CantHandleException(AFixtureException):
    """ if none of the devices is able to handle a cmd_any() call
    """
    pass

class AParseException(AFixtureException):
    """ Base class for exceptions concerning parsing errors.
    """
    pass

class CantParseException(AFixtureException):
    """ is raised when a Fixture cannot parse a given file.
    """
    pass

class NoPropsException(AFixtureException):
    """ is raised when
    """
    pass

class NoDeviceException(AFixtureException):
    """ is raised when a :py:clas:`~monk_tf.fixture.Fixture` requires a device but has none.
    """
    pass

class WrongNameException(AFixtureException):
    """ is raised when no devs with a given name could be found.
    """
    pass

##############################################################
#
# Fixture Classes - creates MONK objects based on dictionaries
#
##############################################################

class Fixture(object):
    """ Creates :term:`MONK` objects based on dictionary like objects.

    This is the class that provides the fundamental feature of this layer. It
    reads data files by trying to parse them via its list of known parsers and
    if it succeeds, it creates :term:`MONK` objects based on the configuration
    given by the data file. Most likely these objects are one or more
    :py:class:`~monk_tf.dev.Device` objects that have at least one
    :py:class:`~monk_tf.conn.AConnection` object each. If more than one
    :term:`fixture file` is read containing the same name on the highest level,
    then the latest data gets used. This does not work on lower levels of
    nesting, though. If you attempt to overwrite lower levels of nesting, what
    actually happens is that the highest layer gets overwritten and you lose
    the data that was stored in the older objects. This is simply how
    :py:meth:`set.update` works.

    One source of data (either a file name or a child class of
    :py:class:`~monk_tf.fixture.AParser`) can be given to an object of this
    class by its constructer, others can be added afterwards with the
    :py:meth:`~monk_tf.fixture.Fixture.read` method. An example looks like
    this::

        import monk_tf.fixture as mf

        fixture = mf.Fixture('/etc/monk_tf/default_devices.cfg')
                .read('~/.monk/default_devices.cfg')
                # can also be a parser object
                .read(XiniParser('~/testsuite12345/suite_devices.cfg'))

    """

    _DEFAULT_CLASSES = {
        "Device" : dev.Device,
        "HydraDevice" : dev.Hydra,
        "SerialConnection" : conn.SerialConn,
        "SshConnection" : conn.SshConn,
    }

    _DEFAULT_DEBUG_SOURCE = "MONK_DEBUG_SOURCE"

    def __init__(self, call_location, name=None, classes=None,
            lookfordbgsrc=True, filename="fixture.cfg", auto_search=True):
        """

        :param call_location: the __file__ from where this is called.

        :param name: The name of this object.

        :param parsers: An :python:term:`iterable` of
                        :py:class:`~monk_tf.fixture.AParser` classes to be used
                        for parsing a given
                        :py:attr:`~monk_tf.fixture.Fixture.source`.

        :param classes: A :py:class:`dict` of classes to class names. Used for
                        parsing the type attribute in
                        :term:`fixture files<fixture file>`.

        :param lookfordbgsrc: If True an environment variable is looked for to
                              read a local debug config. If False it won't be
                              looked for.

        :param filename: the name of the file which contains the configuration.

        :param auto_search: if true, it will automatically search and load
                            fixture files.
        """
        self.call_location = op.dirname(op.abspath(call_location))
        self._logger = logging.getLogger("{}:{}".format(
            __name__,
            name or self.__class__.__name__,
        ))
        self.devs = []
        self._devs_dict = {}
        self.classes = classes or self._DEFAULT_CLASSES
        self.props = config.ConfigObj()
        self.filename = filename
        self.auto_search = auto_search
        # look if the user has a default config in his home dir
        if auto_search:
            self.log("autosearching for fixture files...")
            home_fixture = op.expanduser(op.join("~", self.filename))
            if op.exists(home_fixture):
                self.read(home_fixture)
            # starting from root load all fixtures from parent directories
            self.log("location:{}".format(self.call_location))
            self.log("parent_dirs:" + str(list(self._parent_dirs(self.call_location))))
            for p in reversed(list(self._parent_dirs(self.call_location))):
                fixture_file = op.join(p, self.filename)
                if op.exists(fixture_file):
                    self.read(fixture_file)
        else:
            self.log("auto search deactivated, loaded without looking for fixture files")

    @property
    def name(self):
        return self._logger.name

    @name.setter
    def name(self, new_name):
        self._logger.name = new_name

    def _parent_dirs(self, path):
        """ generate parent directories for path
        """
        while True:
            yield path
            mem = op.dirname(path)
            if path == mem:
                break
            else:
                path = mem


    def read(self, source):
        """ Read more data, either as a file name or as a parser.

        :param source: the data source; either a file name or a
                       :py:class:`~monk_tf.fixture.AParser` child class
                       instance.

        :return: self
        """
        self._logger.debug("read: " + str(source))
        self.tear_down()
        self.props.merge(config.ConfigObj(source))
        self._initialize()
        return self

    def _initialize(self):
        """ Create :term:`MONK` objects based on self's properties.
        """
        self._logger.debug("initialize with props: " + str(self.props))
        if self.props:
            self.devs = [self._parse_section(d, self.props[d]) for d in self.props.viewkeys()]
        else:
            raise NoPropsException("have you created any fixture files?")

    def _parse_section(self, name, section):
        self._logger.debug("parse_section({},{},{})".format(
            str(name),
            type(section).__name__,
            list(section.keys())
        ))
        # TODO section parsing should be wrapped in handlers
        #      so that they can be extended without overwrites
        sectype = self.classes[section.pop("type")]
        if "conns" in section:
            cs = section.pop("conns")
            section["conns"] = [self._parse_section(s, cs[s]) for s in cs]
        if "bcc" in section:
            self.log("DEPRECATED: Use bctrl instead of bcc")
            bs = section.pop("bcc")
            section["bcc"] = self._parse_section("bcc", bs)
        if "bctrl" in section:
            bs = section.pop("bctrl")
            section["bcc"] = self._parse_section("bctrl", bs)
        section["name"] = name
        self.log("load section:" + str(sectype) + "," + str(section))
        return sectype(**section)

    def cmd_first(self, msg, expect=None, timeout=30, login_timeout=None):
        """ call :py:meth:`cmd` from first :py:class:`~monk_tf.device.Device`
        """
        self.log("cmd_first({},{},{},{})".format(
            msg, expect, timeout, login_timeout))
        try:
            return self.devs[0].cmd(msg)
        except IndexError:
            raise NoDeviceException("this fixture has no device loaded")

    def cmd_any(self, msg, expect=None, timeout=30, login_timeout=None):
        self.log("cmd_any({},{},{},{})".format(
            msg, expect, timeout, login_timeout))
        if not self.devs:
            self._logger.warning("fixture has no devices for sending commands to")
        for dev in self.devs:
            try:
                self.log("send cmd '{}' to device '{}'".format(
                    msg.encode("unicode-escape"),
                    dev,
                ))
                return dev.cmd(
                        msg=msg,
                        expect=expect,
                        timeout=timeout,
                        login_timeout=login_timeout,
                )
            except Exception as e:
                self._logger.exception(e)
            raise CantHandleException(
                    "fixt:'{}',devs:{},could not send cmd '{}'".format(
                        self.name,
                        map(str, self.devs),
                        msg.encode('unicode-escape'),
            ))

    def cmd_all(self, msg, expect=None, timeout=30, login_timeout=None):
        self.log("cmd_any({},{},{},{})".format(
            msg, expect, timeout, login_timeout))
        if not self.devs:
            self._logger.warning("fixture has no devices for sending commands to")
        for dev in self.devs:
            self.log("send cmd '{}' to device '{}'".format(
                msg.encode("unicode-escape"),
                dev,
            ))
            return dev.cmd(
                    msg=msg,
                    expect=expect,
                    timeout=timeout,
                    login_timeout=login_timeout,
            )

    def get_dev(self, which):
        try:
            return self.devs[which]
        except TypeError:
            try:
                return self._devs_dict[which]
            except KeyError:
                names = []
                for dev in self.devs:
                    if dev.name == which:
                        self._devs_dict[which] = dev
                        return dev
                    else:
                        names.append(dev.name)
                raise WrongNameException("Couldn't retreive connection with name '{}'. Available names are: {}".format(which, names))

    def reset_config_all(self):
        if not self.devs:
            self._logger.warning("fixture has no devices for sending commands to")
        for dev in self.devs:
            dev.reset_config()


    def log(self, msg):
        self._logger.debug(msg)

    def tear_down(self):
        """ Can be used for explicit destruction of managed objects.

        This should be called in every :term:`test case` as the last step.

        """
        self.log("teardown")
        for device in self.devs:
            device.close_all()
        self.devs = []

    def __str__(self):
        return "{cls}.devs:{devs}".format(
                cls=self.__class__.__name__,
                devs=[str(d) for d in self.devs],
        )

    def __enter__(self):
        self.log("__enter__")
        return [self] + list(self.devs)

    def __exit__(self, exception_type, exception_val, trace):
        self.log("__exit__")
        self.tear_down()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2012, 2013 DResearch Fahrzeugelektronik GmbH
# Maintained by project-monk@dresearch-fe.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
This is the package overview of :term:`MONK`. If anything is unclear, you might
have a look into :ref:`chap-intro`.

The following texts describe the three layers that were explained in
:ref:`intro-layers`:
"""

__title__ = "monk_tf"
__version__ = "0.8.3"
__author__ = "DResearch Fahrzeugelektronik GmbH"
__license__ = "GPLv3+"
__copyright__ = "Copyright 2012, 2013, 2014 DResearch Fahrzeugelektronik GmbH"



# -*- coding: utf-8 -*-
#
# MONK automated test framework
#
# Copyright (C) 2013 DResearch Fahrzeugelektronik GmbH
# Written and maintained by MONK Developers <project-monk@dresearch-fe.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version
# 3 of the License, or (at your option) any later version.
#

"""
This module implements device handling. Using the classes from this module you
can abstract a complete :term:`target device` in a single object. On
instantiation you give it some connections and then (theoretically) let the
device handle the rest.

Example::

    import monk_tf.dev as md
    import monk_tf.conn as mc
    # create a device with a ssh connection and a serial connection
    d=md.Device(
        mc.SshConn('192.168.2.100', 'tester', 'secret'),
        mc.SerialConn('/dev/ttyUSB2', 'root', 'muchmoresecret'),
    )
    # send a command (the same way as with connections)
    print d.cmd('ls -al')
    [...]
"""

import logging
import time
import json

import requests
import pexpect

import conn

logger = logging.getLogger(__name__)

############
#
# Exceptions
#
############

class ADeviceException(Exception):
    """ Base class for exceptions of the device layer.
    """
    pass

class CantHandleException(ADeviceException):
    """ is raised when a request cannot be handled by the connections of a
    :py:class:`~monk_tf.dev.Device`.
    """
    pass

class UpdateFailedException(ADeviceException):
    """ is raised if an update didn't get finished or was rolled back.
    """
    pass

class WrongNameException(ADeviceException):
    """ is raised when no connection with a given name could be found.
    """
    pass


##############################
#
# Devices - currently just one
#
##############################

class Device(object):
    """ is the API abstraction of a :term:`target device`.
    """

    def __init__(self, *args, **kwargs):
        """
        :param conns: list of connections. The following works as well::

            ``Device(OneConnection(...), AnotherConnection(...),...)``

        :param name: Device name for logging purposes.
        """
        self._logger = logging.getLogger(kwargs.pop("name", self.__class__.__name__))
        self.conns = kwargs.pop("conns", list(args))
        self._conns_dict = {}
        self.prompt = PromptReplacement()

    @property
    def name(self):
        return self._logger.name

    @name.setter
    def name(self, new_name):
        self._logger.name = new_name

    def cmd(self, msg, expect=None, timeout=30, login_timeout=None, do_retcode=True):
        """ Send a :term:`shell command` to the :term:`target device`.

        :param msg: the :term:`shell command`.
        :param expect: if you don't expect a prompt in the end but something
                       else, you can add a regex here.
        :param timeout: when command should return without finding what it's
                        looking for in the output. Will raise a
                        :py:exception:`pexpect.Timeout` Exception.
        :param do_retcode: should this command retreive a returncode
        :return: the standard output of the :term:`shell command`.
        """
        self.log("cmd({},{},{},{},{})".format(
            msg, expect, timeout, login_timeout, do_retcode))
        if not self.conns:
            self._logger.warning("device has no connections to use for interaction")
        for connection in self.conns:
            try:
                self.log("send cmd '{}' via connection '{}'".format(
                    msg.encode('unicode-escape'),
                    connection,
                ))
                return connection.cmd(
                        msg=msg,
                        expect=PromptReplacement.replace(connection, expect),
                        timeout=timeout,
                        do_retcode=do_retcode,
                )
            except Exception as e:
                self._logger.exception(e)
        raise CantHandleException(
                "dev:'{}',conns:'{}':could not send cmd '{}'".format(
                    self.name,
                    map(str, self.conns),
                    msg,
        ))

    def get_conn(self, which):
        self.log("get_conn({})".format(which))
        try:
            return self.conns[which]
        except TypeError:
            try:
                return self._conns_dict[which]
            except KeyError:
                names = []
                for conn in self.conns:
                    if conn.name == which:
                        self.log("cache conn in dict:" + which)
                        self._conns_dict[which] = conn
                        return conn
                    else:
                        names.append(conn.name)
                raise WrongNameException("Couldn't retreive connection with name '{}'. Available names are: {}".format(which, names))

    def log(self, msg):
        """ sends a debug-level message to the logger

        This method is used so often, that a smaller version of it is quite
        comfortable.
        """
        self._logger.debug(msg)

    def close_all(self):
        self.log("close_all()")
        for c in self.conns:
            c.close()

    def __str__(self):
        return "{}({}):name={}".format(
                self.__class__.__name__,
                [str(c) for c in self.conns],
                self.name,
        )

class Hydra(Device):
    """ is the device type of DResearch Fahrzeugelektronik GmbH.
    """

    def update(self, link=None, force=None):
        """ update the device to current build from Jenkins.
        """
        self._logger.info("Attempt update to " + str(link or self._update_link))
        if not self.do_update:
            self.log("don't update due to MONK configuration")
            return
        connection_closed = (pexpect.EOF, pexpect.TIMEOUT)
        if not self.is_updated or force:
            _, out = self.cmd("do-update -c && get-update {} && do-update".format(
                    link if link else self._update_link,
                ), expect=("[lL]ogin: ",) +  connection_closed,
                timeout=600,
                do_retcode=False,
            )
            if self.conns[0].exp.after in connection_closed:
                self.log("reset connection after reboot")
                del self.conns[0]._exp
            self.log("wait till device recovered from updating")
            time.sleep(240)
            self.log("continue")
        else:
            self._logger.info("Already updated.")

    def __init__(self, *args, **kwargs):
        self._update_link = "http://hydraip-integration.internal.dresearch-fe.de:8080/view/HIPOS/job/HydraIP_UpdateV3_USB_Stick/lastSuccessfulBuild/artifact/rel-hudson/hyp-updateV3-hikirk.zip"
        self._jenkins_link = "http://hydraip-integration.internal.dresearch-fe.de:8080/view/HIPOS/job/daisy-hipos-dfe-closed-hikirk/api/json"
        # I want BOOLEANS!
        self.do_update = kwargs.pop("update",True) in ("True", True)
        self.do_resetconfig = kwargs.pop("resetconfig",True) in ("True", True)
        super(Hydra, self).__init__(*args, **kwargs)

    @property
    def latest_build(self):
        """ get the latest build ID from jenkins
        """
        out = requests.get(self._jenkins_link).text
        return json.loads(out)["lastSuccessfulBuild"]["number"]

    @property
    def current_fw_version(self):
        """ the current version of the installed firmware
        """
        _, out = self.cmd("do-update --current-update-version | awk '{print $2}'")
        return out

    @property
    def has_newest_firmware(self):
        """ check whether the installed firmware is the newest on jenkins
        """
        return str(self.latest_build) in str(self.current_fw_version)

    @property
    def is_updated(self):
        """ check whether the device is already updated.

        Currently it is implementd with
        :py:meth:`dev.Hydra.has_newest_firmware`.
        """
        return self.has_newest_firmware

    def reset_config(self):
        """ reset the HydraIP configuration on the device
        """
        if not self.do_resetconfig:
            self.log("don't reset config due to MONK configuration")
            return
        # otherwise it might not really be a reset
        self.cmd("rm /etc/drconfig/hydraip.json.good")
        self.cmd(
            msg="rm -rf /var/lib/connman/* && hip-activate-config --reset && sync && halt -p",
            expect=[self.prompt, pexpect.EOF, pexpect.TIMEOUT],
            do_retcode=False,
        )

#########
#
# Helpers
#
#########

class PromptReplacement(object):
    """ should be replaced by each connection's own prompt.
    """
    @classmethod
    def replace(cls, c, expect):
        """ this is an awful workaround...
        """
        if not expect:
            return expect
        if isinstance(expect, str):
            return expect
        if isinstance(expect, Exception):
            return expect
        if not isinstance(expect, list):
            expect = list(expect)
        return [c.prompt if isinstance(e, PromptReplacement) else e for e in expect]
